unit CalcPvd;

{$IFDEF FPC}
  {$MODE Delphi}
{$ENDIF}

interface

uses
  Windows, SysUtils, db,  Variants, Classes,ArtRepProvider;


type TARCalcProvider = class(TArtRepProvider)
 private
    StarOffice: Variant;
    Desktop : Variant;
    Document: Variant;
    ActiveSheet : Variant;
    Ps, Sz, Gp : Variant;
    ifiltname : string;
    ofiltname : string;
    fvisible : boolean;
    fconvert : boolean;
    fcomma : boolean;
    fseparator : boolean;
    fRCconvert : boolean;
    flocalnames : array of string;
    fnames : array of string;
    function PropValue(PropName : variant; PropValue:Variant):variant;
    function Transp(A : Variant; R1,C1,R2,C2:integer) : Variant;
 public
    function Convert(formula:string) : string;
    function ConvertRC(formula:string;  row, col: integer) : string;
    procedure AddConvert(local, native:string);
    procedure Connect; override;
    procedure SetVisible(visible:boolean); override;
    procedure OpenDoc(name:string; page:integer); override;
    procedure CreateDoc(name:string); override;
    procedure SetCalculation(auto:boolean); override;
    function GetCellText(row,col:integer) : string; override;
    procedure SetCellText(row,col:integer; txt:string); override;
    function GetCellFormula(row,col:integer) : string; override;
    procedure SetCellFormula(row,col:integer; Formula:Variant); override;
    procedure DeleteRow(row : integer); override;
    procedure DeleteRows(row, count : integer); override;
    procedure DeleteColumn(col : integer); override;
    procedure InsertColumns(col, count:integer); override;
    procedure CopyColumns(col, tocol, count:integer); override;
    procedure InsertRows(row, count:integer); override;
    procedure CopyRows(row, torow, count:integer); override;
    procedure SetRange(R1,C1,R2,C2:integer; Data:Variant); override;
    procedure AddNameRange(R1,C1,R2,C2:integer; name:string); override;
    procedure AddName(name:string;val:integer); override;
    function LastUsedRow : integer; override;
    function DocLoaded : boolean; override;
    function LastUsedColumn : integer; override;
    procedure SetSheetName(name : string); override;
    procedure Group(row, count : integer); override;
    procedure CleanUp; override;
    procedure AutoFit(col : integer); override;
    procedure ActivateCell(row,col : integer); override;
    procedure ShowLevels(level : integer); override;
    procedure Run(macro : string); override;
    procedure Print(copies:integer=1); override;
    procedure SaveDocAs(s:string); override;
    procedure SaveDoc; override;
    procedure BeforeQuit; override;
    procedure InsertPicture(row,col:integer; filename : string); override;
    procedure CloseDoc; override;
    procedure MarkSaved; override;
    procedure Quit; override;
    function Installed : boolean; override;
    function ProviderName : string; override;
    function GetMyVariant(Field: TField;str:boolean=true): OLEVariant; override;
    function GetDate(val : TDateTime): OLEVariant; override;
    constructor Create(Owner: TComponent); override;
    destructor Destroy; override;
    function ProgLoaded:boolean;
    procedure DoDispatch(Command: Variant; Params: Variant);
  published
    property InputFilterName : string read ifiltname write ifiltname;
    property OutputFilterName : string read ofiltname write ofiltname;
    property ConvertLocalNames : boolean read fconvert write fconvert default false;
    property ConvertDecimal : boolean read fseparator write fseparator default false;
    property ConvertComma : boolean read fcomma write fcomma default true;
    property ConvertRCRef : boolean read fRCconvert write fRCconvert default true;
end;

implementation

uses oleserver, comobj, dateutils  {$IFDEF FPC}, Fileutil  {$ENDIF};

constructor TARCalcProvider.Create(Owner: TComponent);
begin
  inherited Create(Owner);
  //ifiltname := 'MS Excel 97 Vorlage/Template';
  ifiltname := 'MS Excel 97';
  ofiltname := 'MS Excel 97';
  fRCconvert := true;
  fcomma := true;
end;

destructor TARCalcProvider.Destroy;
begin
  inherited Destroy;
end;

procedure TARCalcProvider.Connect;
begin
   if not ProgLoaded then
      StarOffice := CreateOleObject('com.sun.star.ServiceManager');
   if ProgLoaded then
   begin
    Desktop := StarOffice.createInstance('com.sun.star.frame.Desktop');
    Ps := StarOffice.Bridge_GetStruct('com.sun.star.awt.Point');
    Sz := StarOffice.Bridge_GetStruct('com.sun.star.awt.Size');
    GP:=StarOffice.CreateInstance('com.sun.star.graphic.GraphicProvider');
   end;
end;

procedure TARCalcProvider.SetSheetName(name : string);
begin
 {$IFDEF FPC}
   ActiveSheet.setName(widestring(Name));
 {$ELSE}
   ActiveSheet.setName(Name);
 {$ENDIF}
end;

function TARCalcProvider.Installed : boolean;
begin
 try
  StarOffice := CreateOleObject('com.sun.star.ServiceManager');
 except
 end; 
 result := ProgLoaded;
end;

function TARCalcProvider.ProviderName : string;
begin
 result := 'OpenOffice Calc';
end;


procedure TARCalcProvider.SetVisible(visible:boolean);
begin
  if Docloaded then
  begin
   if visible then
   begin
    Document.unLockControllers;
    Document.removeActionLock;
   end
   else
   begin
    Document.lockControllers;
    Document.addActionLock;
   end;
   Document.getCurrentController.getFrame.getContainerWindow.setVisible(visible);
  end;
  fvisible := visible
end;

procedure TARCalcProvider.AddConvert(local, native:string);
begin
 setlength(flocalnames, high(flocalnames)+2);
 setlength(fnames, high(fnames)+2);
 flocalnames[high(flocalnames)] := AnsiUpperCase(local);
 fnames[high(fnames)] := AnsiUpperCase(native);
end;

function FileNameToURL(FileName: string): {$IFDEF FPC}widestring {$ELSE} string {$ENDIF};
begin
  result:= '';
  filename := {$IFDEF FPC} ExpandFileNameUTF8(Filename); { *Converted from ExpandFileName*  }
                   {$ELSE}  ExpandFileName(Filename); {$ENDIF}

  if LowerCase(copy(FileName,1,8))<>'file:///' then
    result:= 'file:///';
  result:= result + StringReplace(FileName, '\', '/', [rfReplaceAll, rfIgnoreCase]);
  result := StringReplace(result, ' ', '%20', [rfReplaceAll, rfIgnoreCase]);
end;


procedure TARCalcProvider.OpenDoc(name:string; page:integer);
var
   VariantArr: variant;
begin
   VariantArr :=    VarArrayCreate([0, 2], varVariant);
   VariantArr[0] := PropValue('Hidden', not fvisible);
   VariantArr[1] := PropValue('FilterName', ifiltname);
   VariantArr[2] := PropValue('MacroExecutionMode',4); // ALWAYS_EXECUTE_NO_WARN
   Document := Desktop.LoadComponentFromURL(
                  FileNameToURL(Name), '_blank', 0,
                  VariantArr);
   if page > 0 then Document.GetCurrentController.SetActiveSheet(Document.GetSheets.GetByIndex(page-1));
   ActiveSheet := Document.GetCurrentController.GetActiveSheet;
end;

procedure TARCalcProvider.CreateDoc(name:string);
var
   VariantArr: variant;
begin
   VariantArr :=    VarArrayCreate([0, 3], varVariant);
   VariantArr[0] := PropValue('Hidden', not fVisible);
   VariantArr[1] := PropValue('FilterName', ifiltname);
   VariantArr[2] := PropValue('AsTemplate',true);
   VariantArr[3] := PropValue('MacroExecutionMode',4); // ALWAYS_EXECUTE_NO_WARN

   Document := Desktop.LoadComponentFromURL(
                  FileNameToURL(Name), '_blank', 0,
                  VariantArr);
   ActiveSheet := Document.GetCurrentController.GetActiveSheet;
end;

procedure TARCalcProvider.SetCalculation(auto:boolean);
begin
   Document.enableAutomaticCalculation(auto);
end;

function TARCalcProvider.GetCellText(row,col:integer) : string;
begin
  result:= ActiveSheet.getCellByPosition(col-1, row-1).getstring;
end;

Procedure TARCalcProvider.SetCellText(row,col:integer; txt:string);
begin
 {$IFDEF FPC}
  ActiveSheet.getCellByPosition(col-1, row-1).setstring(widestring(txt));
 {$ELSE}
  ActiveSheet.getCellByPosition(col-1, row-1).setstring(txt);
 {$ENDIF}
end;


procedure TARCalcProvider.InsertPicture(row,col:integer; filename : string);
var sh,cr,cell,pic, url: variant;
    i,x,y,sx,sy,hx,hy : integer;
begin
 try
   cell := ActiveSheet.GetCellByPosition(col-1,row-1);
   cr := ActiveSheet.CreateCursorByRange(cell);
   sx := 0;
   sy := 0;
   Cr.CollapseToMergedArea;
   for i := Cr.RangeAddress.StartColumn to Cr.RangeAddress.EndColumn do
     sx := sx + ActiveSheet.GetCellByPosition(i,row-1).Size.width;
   for i := Cr.RangeAddress.StartRow to Cr.RangeAddress.EndRow do
     sy := sy + ActiveSheet.GetCellByPosition(col-1,i).Size.height;
   Url := VarArrayCreate([0, 0], varVariant);
   Url[0]:=PropValue('URL',FileNameToURL(filename));
   pic := GP.QueryGraphic(Url);
   x := pic.Size.Width;
   y := pic.Size.Height;

   hx := sx;
   hy := sx * y div x;
   if hy > sy then
   begin
     hy := sy;
     hx := sy * x div y;
   end;

   Sz.Width:=hx;
   Sz.Height:=hy;

   Ps.X:=cell.Position.X + (sx - hx) div 2;
   Ps.Y:=cell.Position.Y + (sy - hy) div 2;


   Sh:=Document.CreateInstance('com.sun.star.drawing.GraphicObjectShape');
   Sh.Position:=Ps;
   Sh.Size:=Sz;
   Sh.Graphic:=pic;
   ActiveSheet.DrawPage.Add(Sh);
 except
 end;
   sh := UnAssigned;
   url := UnAssigned;
   cr := UnAssigned;
   cell := UnAssigned;
   pic := UnAssigned;
end;

const nonalpha = ['`','~','!','@','#','$','%','^','&','*','(',')','-','+','=','\','|',']','}','[','{',';',':','''','<',',','>','.','/','?'];
const digits = ['0'..'9'];

function TARCalcProvider.ConvertRC(formula:string; row, col: integer) : string;
var p,dr,dc,mdr,mdc:integer; name,s: string; refr, refc : boolean;
begin
 result := '';
 p := 0;
 while p<length(formula) do
 begin
  inc(p);
  if formula[p] in nonalpha then
  begin
   if (formula[p] = ',') and fcomma then result := result + ';'
   else result := result+formula[p];
   continue;
  end;
  if formula[p]='"' then
  begin
   result := result+formula[p];
   inc(p);
   while (p<=length(formula)) and (formula[p] <> '"') do
   begin
    result := result+formula[p];
    inc(p);
   end;
   if p <= length(formula) then result := result+'"';
   continue;
  end;
  if formula[p] in digits then
  begin
   result := result+formula[p];
   inc(p);
   while (p<=length(formula)) and (formula[p] in digits) do
   begin
    result := result+formula[p];
    inc(p);
   end;
   dec(p);
   continue;
  end;
  if (formula[p] <> 'r') and (formula[p] <> 'R') then
  begin
    while (p<=length(formula)) and (not ((formula[p] in nonalpha) or (formula[p] in digits) or (formula[p] = '"'))) do
    begin
     result := result+formula[p];
     inc(p);
    end;
    dec(p);
  end
  else
  begin
    name := formula[p];
    inc(p);
    dr := 0;
    dc := 0;
    mdr := 1;
    mdc := 1;
    refr := true;
    refc := true;
    if  formula[p] = '[' then
    begin
     name := name + formula[p];
     inc(p);
     if formula[p] = '-' then
     begin
       name := name + formula[p];
       inc(p);
       mdr := -1;
     end;
     s := '';
     while (p<length(formula)) and (formula[p] in digits) do
     begin
       name := name + formula[p];
       s := s +  formula[p];
       inc(p);
     end;
     try
      dr := mdr * strtoint(s);
     except
      result := result + name;
      dec(p);
      continue;
     end;
     if  formula[p] <> ']' then
     begin
      result := result + name;
      dec(p);
      continue;
     end;
     name := name + formula[p];
     inc(p);
    end
    else
    begin
     if (formula[p] <> 'c') and (formula[p] <> 'C') then
     begin
      s := '';
      while (p<length(formula)) and (formula[p] in digits) do
      begin
        name := name + formula[p];
        s := s +  formula[p];
        inc(p);
      end;

      if (s='') and (p<=length(formula)) and (((formula[p] in nonalpha) or (formula[p]='c') or (formula[p]='C')  or (formula[p] = '"'))) then
        dr := row
      else
      try
       dr := strtoint(s);
      except
       result := result + name;
       dec(p);
       continue;
      end;
      refr := false;
     end;
    end;


    if (formula[p] <> 'c') and (formula[p] <> 'C') then
    begin
     result := result + name;
     dec(p);
     continue;
    end;
    name := name + formula[p];
    inc(p);
    if  formula[p] = '[' then
    begin
     name := name + formula[p];
     inc(p);
     if formula[p] = '-' then
     begin
       name := name + formula[p];
       inc(p);
       mdc := -1;
     end;
     s := '';
     while (p<length(formula)) and (formula[p] in digits) do
     begin
       name := name + formula[p];
       s := s +  formula[p];
       inc(p);
     end;
     try
      dc := mdc * strtoint(s);
     except
      result := result + name;
      dec(p);
      continue;
     end;
     if  formula[p] <> ']' then
     begin
      result := result + name;
      dec(p);
      continue;
     end;
     name := name + formula[p];
     inc(p);
    end
    else
    begin
     s := '';
     while (p<length(formula)) and (formula[p] in digits) do
     begin
       name := name + formula[p];
       s := s +  formula[p];
       inc(p);
     end;
     if (s='') and (p<=length(formula)) and (((formula[p] in nonalpha) or (formula[p] = '"'))) then
     begin
      dc := COL;
     end
     else
     try
      dc := strtoint(s);
     except
      result := result + name;
      dec(p);
      continue;
     end;
     refc := false;
    end;

    if (p>=length(formula)) or (formula[p] in nonalpha) then
    begin
      if refc then result := result + colname(col+dc)
      else result := result + colname(dc);
      if refr then result := result + inttostr(row+dr)
      else result := result + inttostr(dr);
    end
    else
      result := result + name;
    dec(p);
  end;
 end;
end;


function TARCalcProvider.Convert(formula:string) : string;
var p,i:integer; name, name1: string; found : boolean;
begin
 result := '';
 p := 0;
 while p<length(formula) do
 begin
  inc(p);
  if formula[p] in nonalpha then
  begin
   if (formula[p] = DecimalSeparator) and fseparator then result := result + '.'
   else result := result+formula[p];
   continue;
  end;
  if formula[p]='"' then
  begin
   result := result+formula[p];
   inc(p);
   while (p<=length(formula)) and (formula[p] <> '"') do
   begin
    result := result+formula[p];
    inc(p);
   end;
   if p <= length(formula) then result := result+'"';
   continue;
  end;
  if formula[p] in digits then
  begin
   result := result+formula[p];
   inc(p);
   while (p<=length(formula)) and (formula[p] in digits) do
   begin
    result := result+formula[p];
    inc(p);
   end;
   dec(p);
   continue;
  end;
  name := '';
  while (p<=length(formula)) and (not ((formula[p] in nonalpha) or (formula[p] = '"')) or (formula[p]='.')) do
  begin
   name := name + formula[p];
   inc(p);
  end;
  found := false;
  name1 := AnsiUpperCase(name);
  for i:=0 to high(flocalnames) do
   if name1 = flocalnames[i] then
   begin
    result := result+fnames[i];
    found := true;
    break;
   end;
  if not found then result := result + name;
  dec(p);
 end;
end;

function TARCalcProvider.GetCellFormula(row,col:integer) : string;
begin
   result:= ActiveSheet.getCellByPosition(col-1, row-1).FormulaLocal;
  if fconvert then result := convert(result);
end;

Procedure TARCalcProvider.SetCellFormula(row,col:integer; Formula:Variant);
var s:string;
begin
// StarOffice.DefaultContext.getValueByName('/singletons/org.openoffice.vba.theGlobals').Application.Range(colname(col)+inttostr(row),colname(col)+inttostr(row)).FormulaR1C1 := Formula;
  if fRCconvert and VarIsstr(Formula) then
   begin
    s := Formula;
    if (length(s) > 0) and (s[1]='=') then

   {$IFDEF FPC}
    ActiveSheet.getCellByPosition(col-1, row-1).Formula  := widestring(ConvertRC(s,row,col))
    else
     ActiveSheet.getCellByPosition(col-1, row-1).Formula := widestring(s);
    {$ELSE}
    ActiveSheet.getCellByPosition(col-1, row-1).Formula  := ConvertRC(s,row,col)
    else
     ActiveSheet.getCellByPosition(col-1, row-1).Formula := s;
    {$ENDIF}

   end
   else
     if VarIsNumeric(Formula) then
        ActiveSheet.getCellByPosition(col-1, row-1).Value := Formula
     else
        ActiveSheet.getCellByPosition(col-1, row-1).Formula := Formula;
end;

procedure TARCalcProvider.DeleteRow(row : integer);
begin
  ActiveSheet.getRows.removeByIndex(Row-1, 1);
end;

procedure TARCalcProvider.DeleteColumn(col : integer);
begin
  ActiveSheet.getColumns.removeByIndex(Col-1, 1);
end;

procedure TARCalcProvider.InsertColumns(col, count:integer);
begin
  ActiveSheet.getColumns.insertByIndex(col - 1, count);
end;

procedure TARCalcProvider.CopyColumns(col, tocol, count:integer);
begin
 {$IFDEF FPC}
  ActiveSheet.CopyRange(ActiveSheet.getCellRangeByName(widestring(colname(ToCol)+'1:'+colname(ToCol+count-1)+'65536')).getCellByPosition(0,0).CellAddress,
//          ActiveSheet.getCellRangeByName(widestring(colname(Col)+'1:'+colname(Col+count-1)+'65536')).RangeAddress);
          ActiveSheet.getCellRangeByPosition(col-1,0,col-1+count-1,ActiveSheet.RangeAddress.endrow).RangeAddress);
 {$ELSE}
  ActiveSheet.CopyRange(ActiveSheet.getCellRangeByName(colname(ToCol)+'1:'+colname(ToCol+count-1)+'65536').getCellByPosition(0,0).CellAddress,
//          ActiveSheet.getCellRangeByName(colname(Col)+'1:'+colname(Col+count-1)+'65536').RangeAddress);
          ActiveSheet.getCellRangeByPosition(col-1,0,col-1+count-1,ActiveSheet.RangeAddress.endrow).RangeAddress);
 {$ENDIF}
end;


procedure TARCalcProvider.InsertRows(row, count:integer);
begin
 ActiveSheet.getRows.insertByIndex(Row - 1, count);
end;

procedure TARCalcProvider.DeleteRows(row, count:integer);
begin
  ActiveSheet.getRows.removeByIndex(Row-1, count);
end;


procedure TARCalcProvider.CopyRows(row, torow, count:integer);
begin
 {$IFDEF FPC}
// ActiveSheet.CopyRange(ActiveSheet.getCellRangeByName(widestring('A'+inttostr(Torow))).getCellByPosition(0,0).CellAddress,
//          ActiveSheet.getCellRangeByName(widestring('A'+inttostr(row)+':IV'+inttostr(row+count-1))).RangeAddress);
 ActiveSheet.CopyRange(ActiveSheet.getCellRangeByName(widestring('A'+inttostr(Torow))).getCellByPosition(0,0).CellAddress,
          ActiveSheet.getCellRangeByPosition(0,row-1,ActiveSheet.RangeAddress.endcolumn,row-1+count-1).RangeAddress);
 {$ELSE}

 ActiveSheet.CopyRange(ActiveSheet.getCellRangeByName('A'+inttostr(Torow)).getCellByPosition(0,0).CellAddress,
          ActiveSheet.getCellRangeByPosition(0,row-1,ActiveSheet.RangeAddress.endcolumn,row-1+count-1).RangeAddress);
{$ENDIF}
end;

function TARCalcProvider.Transp(A : Variant; R1,C1,R2,C2:integer) : Variant;
var i,j:integer; s:string;
begin
 Result := VarArrayCreate([1,C2-C1+1,1,R2-R1+1],varVariant);
 for i:=1 to C2-C1+1 do
  for j:=1 to R2-R1+1 do
  begin
   if fRCconvert and VarIsstr(A[j,i]) then
   begin
    s := A[j,i];
    if (length(s) > 0) and (s[1]='=') then
     Result[i,j] := ConvertRC(s,r1+j-1,c1+i-1)
    else
     Result[i,j] := s;
   end
   else
    Result[i,j] := A[j,i];

  end;
end;

procedure TARCalcProvider.SetRange(R1,C1,R2,C2:integer; Data:Variant);
var Cells, TrData:Variant;
begin
 {$IFDEF FPC}
  Cells := ActiveSheet.getCellRangeByName(widestring(colname(C1)+inttostr(R1)+':'+colname(C2)+inttostr(R2)));
 {$ELSE}
  Cells := ActiveSheet.getCellRangeByName(colname(C1)+inttostr(R1)+':'+colname(C2)+inttostr(R2));
 {$ENDIF}
 TrData := Transp(Data,R1,C1,R2,C2);
 Cells.setFormulaArray(TrData);
 TrData := Unassigned;
 Cells := Unassigned;
end;

procedure TARCalcProvider.AddNameRange(R1,C1,R2,C2:integer; name:string);
var Range:widestring;
begin
//  if pos(' ',ActiveSheet.getName) > 0 then
   Range := '$'''+ActiveSheet.getName+'''.$'+colname(C1)+'$'+inttostr(R1)+':$'+colname(C2)+'$'+inttostr(R2);
//  else
//   Range := '$'+ActiveSheet.getName+'.$'+colname(C1)+'$'+inttostr(R1)+':$'+colname(C2)+'$'+inttostr(R2);
  {$IFDEF FPC}
  if Document.DatabaseRanges.HasByName(widestring(Name)) then
     Document.DatabaseRanges.RemoveByName(widestring(Name));
  if Document.NamedRanges.HasByName(widestring(Name)) then
     Document.NamedRanges.RemoveByName(widestring(Name));
  Document.DatabaseRanges.addNewByName(widestring(Name),ActiveSheet.getCellRangeByName(Range).RangeAddress);
  {$ELSE}
  if Document.DatabaseRanges.HasByName(Name) then
     Document.DatabaseRanges.RemoveByName(Name);
  if Document.NamedRanges.HasByName(Name) then
     Document.NamedRanges.RemoveByName(Name);
//  Document.DatabaseRanges.addNewByName(Name,ActiveSheet.getCellRangeByName(Range).RangeAddress);
  Document.NamedRanges.addNewByName(Name,Range,ActiveSheet.getCellRangeByName(Range).getCellByPosition(0,0).CellAddress,0);
  //Document.NamedRanges.addNewByName(Name,Range,StarOffice.Bridge_GetStruct('com.sun.star.table.CellAddress'),0);
  {$ENDIF}

end;


procedure TARCalcProvider.AddName(name:string;val:integer);
begin
 {$IFDEF FPC}
 if Document.NamedRanges.HasByName(widestring(Name)) then
    Document.NamedRanges.RemoveByName(widestring(Name));
 Document.NamedRanges.addNewByName(widestring(Name),val,ActiveSheet.getCellByPosition(0,0).CellAddress,0);
 {$ELSE}
 if Document.NamedRanges.HasByName(Name) then
    Document.NamedRanges.RemoveByName(Name);
 Document.NamedRanges.addNewByName(Name,Val,ActiveSheet.getCellByPosition(0,0).CellAddress,0);
 {$ENDIF}
end;


function TARCalcProvider.LastUsedRow : integer;
var Cursor : Variant;
begin
  Cursor := ActiveSheet.createCursorByRange(ActiveSheet.GetCellbyPosition(0, 0));
  Cursor.GotoEndOfUsedArea(True);
  Result := Cursor.RangeAddress.EndRow + 1;
  Cursor := Unassigned;
end;

function TARCalcProvider.LastUsedColumn : integer;
var Cursor : Variant;
begin
  Cursor := ActiveSheet.createCursorByRange(ActiveSheet.GetCellbyPosition(0, 0));
  Cursor.GotoEndOfUsedArea(True);
  Result := Cursor.RangeAddress.EndColumn + 1;
  Cursor := Unassigned;
end;

procedure TARCalcProvider.CleanUp;
begin
  Document.unLockControllers;
  Document.removeActionLock;
  SetCalculation(true);
  Document.CalculateAll;
end;

procedure TARCalcProvider.AutoFit(col : integer);
begin
 ActiveSheet.getColumns.getByIndex(col-1).optimalWidth := true;
end;

procedure TARCalcProvider.ActivateCell(row,col : integer);
begin
  Document.CurrentController.Select(ActiveSheet.GetCellbyPosition(col-1,row-1))
end;

procedure TARCalcProvider.ShowLevels(level : integer);
begin
  ActiveSheet.showlevel(level-1,1);
end;

procedure TARCalcProvider.Run(macro : string);
begin
  DoDispatch('macro://./'+macro, VarArrayCreate([0, -1], varVariant));
end;

procedure TARCalcProvider.Print;
var  VariantArr: variant;
begin
    //NOTE: OpenOffice will print all sheets with Printable areas, but if no
    //printable areas are defined in the doc, it will print all entire sheets.
    //Optional parameters (wait until fully sent to printer)...
    SetVisible(true);
    VariantArr:=  VarArrayCreate([0, 2], varVariant);
    VariantArr[0]:= PropValue('Wait', true);
    if copies < 1 then copies := 1;
    VariantArr[1]:= PropValue('CopyCount',copies);
    VariantArr[2]:= PropValue('Collate',true);
    Document.Print(VariantArr);
    VariantArr := Unassigned;
end;

procedure TARCalcProvider.SaveDocAs(s:string);
var
   VariantArr: variant;
begin
   VariantArr := VarArrayCreate([0, 0], varVariant);
   VariantArr[0] := PropValue('FilterName', ofiltname);
   //VariantArr[0] :=  PropValue('Overwrite', True);
   Document.StoreToURL(FileNameToURL(s), VariantArr);
end;

procedure TARCalcProvider.SaveDoc;
begin
  Document.store;
end;

procedure TARCalcProvider.BeforeQuit;
begin
end;

procedure TARCalcProvider.CloseDoc;
begin
  Document.Close(True);
  Document:=Unassigned;
end;

procedure TARCalcProvider.MarkSaved;
begin
 Document.setModified(False);
end;

procedure TARCalcProvider.Quit;
begin
   Desktop.terminate;
   Sz := Unassigned;
   Ps := Unassigned;
   GP := Unassigned;
   StarOffice := Unassigned;
   Desktop := Unassigned;
//   Document := null;
//   Activesheet := null;
end;

function TARCalcProvider.GetDate(val : TDateTime): OLEVariant;
begin
 Result := formatdatetime('mm"/"dd"/"yyyy hh:nn:ss', val);
{ Result :=  '=DATE('+inttostr(yearof(val))+';'+inttostr(monthof(val))+
     ';'+inttostr(dayof(val))+')+TIME('+inttostr(hourof(val))+';'+
      inttostr(minuteof(val))+';'+inttostr(secondof(val))+')';
} 
end;


function TARCalcProvider.GetMyVariant(Field: TField;str:boolean): OLEVariant;
var c:char;
begin
  case Field.DataType of
    ftString, ftFixedChar, ftWideString, ftMemo, ftFmtMemo:
    if str then
      if Field.AsString <> '' then Result := ''''+Field.AsString else Result := ''
    else
     Result := Field.AsString;
    ftSmallint, ftInteger, ftWord, ftLargeint, ftAutoInc: Result := Field.AsInteger;
    ftFloat, ftCurrency, ftBCD:
    begin
      c := decimalseparator;
      if str then decimalseparator := '.';
      Result := formatfloat('0.############################',Field.asfloat);
      decimalseparator := c;
    end;
    ftBoolean: Result := Field.AsBoolean;
    ftDate, ftDateTime:
     if str then Result := GetDate(Field.AsDateTime)
     else Result := DateTostr(Field.AsDateTime);
    ftTime:
     Result := Formatdatetime('t',Field.AsFloat);
  end;
  if VarIsNull(Result) then  Result := '';
end;

procedure TARCalcProvider.Group(row, count : integer);
begin
 {$IFDEF FPC}
   ActiveSheet.group(ActiveSheet.getCellRangeByName(widestring('A'+inttostr(row)+':IV'+inttostr(row+count - 1))).RangeAddress,1);
 {$ELSE}
   ActiveSheet.group(ActiveSheet.getCellRangeByName('A'+inttostr(row)+':IV'+inttostr(row+count - 1)).RangeAddress,1);
 {$ENDIF}
end;

function TARCalcProvider.ProgLoaded : boolean;
begin
 result := not (VarIsEmpty(StarOffice) or VarIsNull(StarOffice));
end;

function TARCalcProvider.DocLoaded : boolean;
begin
 result := ProgLoaded and not (VarIsEmpty(Document) or VarIsNull(Document));
end;

procedure TARCalcProvider.DoDispatch(Command: variant; Params: variant);
var
  Dispatcher, Frame: variant;
begin
    if (VarIsEmpty(Params) or VarIsNull(Params)) then
      Params:= VarArrayCreate([0, -1], varVariant);
    Frame:= Document.getCurrentController.getFrame;
    Dispatcher:= StarOffice.createInstance('com.sun.star.frame.DispatchHelper');
    Dispatcher.executeDispatch(Frame, Command, '', 0, Params);
end;

function TARCalcProvider.PropValue(PropName:variant; PropValue:Variant):variant;
begin
    Result := StarOffice.Bridge_GetStruct('com.sun.star.beans.PropertyValue');
    Result.Name := PropName;
    Result.Value := PropValue;
end;

end.
