unit ExcelPvd;
interface

uses
  Windows, SysUtils, db,  Variants, Classes, Excel2000,artrepprovider;

type
 range = excelrange;

type TARExcelProvider = class(TArtRepProvider)
 private
   EA : TExcelApplication;
   WBk: _Workbook;
   WS : _WorkSheet;
   lcid : integer;
   flocal : boolean;
 public
    procedure Connect; override;
    procedure SetVisible(visible:boolean); override;
    procedure OpenDoc(name:string; page:integer); override;
    procedure CreateDoc(name:string); override;
    procedure SetCalculation(auto:boolean); override;
    function GetCellText(row,col:integer) : string; override;
    procedure SetCellText(row,col:integer; txt:string); override;
    function GetCellFormula(row,col:integer) : string; override;
    procedure SetCellFormula(row,col:integer; Formula:Variant); override;
    procedure DeleteRow(row : integer); override;
    procedure DeleteRows(row, count : integer); override;
    procedure DeleteColumn(col : integer); override;
    procedure InsertColumns(col, count:integer); override;
    procedure CopyColumns(col, tocol, count:integer); override;
    procedure InsertRows(row, count:integer); override;
    procedure CopyRows(row, torow, count:integer); override;
    procedure SetRange(R1,C1,R2,C2:integer; Data:Variant); override;
    procedure AddNameRange(R1,C1,R2,C2:integer; name:string); override;
    procedure AddName(name:string;val:integer); override;
    procedure SetSheetName(name : string); override;
    function LastUsedRow : integer; override;
    function DocLoaded : boolean; override;
    function LastUsedColumn : integer; override;
    procedure Group(row, count : integer); override;
    procedure CleanUp; override;
    procedure AutoFit(col : integer); override;
    procedure ActivateCell(row,col : integer); override;
    procedure ShowLevels(level : integer); override;
    procedure Run(macro : string); override;
    procedure Print; override;
    procedure SaveDocAs(s:string); override;
    procedure SaveDoc; override;
    procedure BeforeQuit; override;
    procedure CloseDoc; override;
    procedure MarkSaved; override;
    procedure Quit; override;
    function Installed : boolean; override;
    function ProviderName : string; override;
    function GetMyVariant(Field: TField;str:boolean=true): OLEVariant; override;
    function GetDate(val : TDateTime): OLEVariant; override;
    constructor Create(Owner: TComponent); override;
    destructor Destroy; override;
  published
    property FormulaLocal : boolean read flocal write flocal default false;   
end;

implementation

uses oleserver, activex;

constructor TARExcelProvider.Create(Owner: TComponent);
begin
  inherited Create(Owner);
  lcid := LOCALE_USER_DEFAULT;//GetUserDefaultLCID;
end;

destructor TARExcelProvider.Destroy;
begin
  Freeandnil(ea);
  inherited Destroy;
end;

procedure TARExcelProvider.Connect;
begin
  EA := TExcelApplication.Create(Self);
  EA.ConnectKind := ckNewInstance;
  EA.AutoConnect := false;
  EA.AutoQuit := false;
  EA.Connect;
end;

procedure TARExcelProvider.SetSheetName(name : string);
begin
 ws.Name := name;
end;

function TARExcelProvider.Installed : boolean;
var ClassID: TCLSID;
begin
  result := CLSIDFromProgID(PWideChar(WideString('Excel.Application')), ClassID) = S_OK;
end;

function TARExcelProvider.ProviderName : string;
begin
 result := 'MS Excel';
end;

procedure TARExcelProvider.SetVisible(visible:boolean);
begin
   ea.ScreenUpdating[lcid] := visible;
   ea.EnableEvents := visible;
   ea.Visible[lcid] := visible;
end;

procedure TARExcelProvider.OpenDoc(name:string; page:integer);
begin
   wbk := ea.Workbooks.open(name,emptyparam,emptyparam,emptyparam,
                emptyparam,emptyparam,emptyparam,emptyparam,
                emptyparam,emptyparam,emptyparam,emptyparam,
                emptyparam,lcid);
   if page > 0 then WS := Wbk.Sheets.item[page] as _WorkSheet
   else WS := WBk.ActiveSheet as _WorkSheet;
   WS.Activate(lcid);
end;

procedure TARExcelProvider.CreateDoc(name:string);
begin
   wbk := ea.Workbooks.Add(name,lcid);
   WS := WBk.ActiveSheet as _WorkSheet;
end;

procedure TARExcelProvider.SetCalculation(auto:boolean);
begin
  if not auto then
   ea.Calculation[lcid] := xlCalculationManual
  else
    ea.Calculation[lcid] := xlCalculationAutomatic;

end;

function TARExcelProvider.GetCellText(row,col:integer) : string;
begin
  result := ws.Cells.Item[row,col].text;
end;

Procedure TARExcelProvider.SetCellText(row,col:integer; txt:string);
begin
 ws.Cells.Item[row,col].value := txt;
end;

function TARExcelProvider.GetCellFormula(row,col:integer) : string;
begin
 if flocal then
  result := ws.Cells.Item[row,col].FormulaR1C1Local
 else
  result := ws.Cells.Item[row,col].FormulaR1C1;
end;

Procedure TARExcelProvider.SetCellFormula(row,col:integer; Formula:Variant);
begin
 if flocal then
  ws.Cells.Item[row,col].FormulaR1C1Local := Formula
 else
  ws.Cells.Item[row,col].FormulaR1C1 := Formula
end;


procedure TARExcelProvider.DeleteRow(row : integer);
begin
 ws.Cells.item[row,1].EntireRow.delete;
end;

procedure TARExcelProvider.DeleteColumn(col : integer);
begin
  ws.Cells.Item[1,col].EntireColumn.Delete;
end;

procedure TARExcelProvider.InsertColumns(col, count:integer);
begin
 ws.range[colname(col)+'1',colname(col+count-1)+'1'].Entirecolumn.Insert(xlshifttoright);
end;

procedure TARExcelProvider.CopyColumns(col, tocol, count:integer);
begin
 ws.range[colname(col)+'1',colname(col+count-1)+'1'].EntireColumn.copy(ws.range[colname(tocol)+'1',colname(tocol+count-1)+'1'].Entirecolumn);
end;


procedure TARExcelProvider.InsertRows(row, count:integer);
begin
 ws.range['A'+inttostr(row),'A'+inttostr(row+count-1)].EntireRow.Insert(xlshiftDown);
end;

procedure TARExcelProvider.DeleteRows(row, count:integer);
begin
 ws.range['A'+inttostr(row),'A'+inttostr(row+count-1)].EntireRow.delete(emptyparam);
end;


procedure TARExcelProvider.CopyRows(row, torow, count:integer);
begin
 ws.range['A'+inttostr(row),'A'+inttostr(row+count-1)].EntireRow.copy(ws.range['A'+inttostr(torow),'A'+inttostr(torow)].Entirerow);
end;

procedure TARExcelProvider.SetRange(R1,C1,R2,C2:integer; Data:Variant);
begin
 if flocal then
  ws.Range[colname(C1)+inttostr(R1),colname(C2)+inttostr(R2)].FormulaR1C1Local := Data
 else
  ws.Range[colname(C1)+inttostr(R1),colname(C2)+inttostr(R2)].FormulaR1C1 := Data;
end;

procedure TARExcelProvider.AddNameRange(R1,C1,R2,C2:integer; name:string); 
begin
  wbk.Names.Add(''''+ws.Name+'''!'+name,EmptyParam,EmptyParam,EmptyParam,EmptyParam,
          EmptyParam,EmptyParam,EmptyParam,EmptyParam,
          '='''+ws.Name+'''!R'+inttostr(R1)+'C'+inttostr(C1)+':R'+inttostr(R2)+'C'+inttostr(C2)
          ,EmptyParam);
end;


procedure TARExcelProvider.AddName(name:string;val:integer);
begin
  wbk.Names.Add(''''+ws.Name+'''!'+name,EmptyParam,EmptyParam,EmptyParam,EmptyParam,
          EmptyParam,EmptyParam,EmptyParam,EmptyParam,'='+inttostr(val),EmptyParam);
end;
  

function TARExcelProvider.LastUsedRow : integer;
begin
 result := ws.UsedRange[lcid].rows.Count;
end;

function TARExcelProvider.LastUsedColumn : integer;
begin
 result := ws.UsedRange[lcid].columns.Count;
end;

procedure TARExcelProvider.CleanUp; 
begin
  ea.ActiveWindow.LargeScroll(emptyparam,1000,emptyparam,emptyparam);
  ea.Calculation[lcid] := xlCalculationAutomatic;
  ea.ScreenUpdating[lcid] := true;
  ea.EnableEvents := true;
  ws.Calculate(lcid);
end;

procedure TARExcelProvider.AutoFit(col : integer); 
begin
 ws.Cells.Item[1,col].entirecolumn.autofit;
end; 

procedure TARExcelProvider.ActivateCell(row,col : integer); 
begin
  ws.Cells.Item[row,col].Activate;
end; 

procedure TARExcelProvider.ShowLevels(level : integer);
begin
 ws.Outline.ShowLevels(level,emptyparam);
end;

procedure TARExcelProvider.Run(macro : string); 
begin
 ea.Run(macro);
end; 

procedure TARExcelProvider.Print; 
begin
   ws.PrintOut(emptyparam,emptyparam,emptyparam,emptyparam,emptyparam, emptyparam,emptyparam,emptyparam,lcid);
end;

procedure TARExcelProvider.SaveDocAs(s:string); 
begin
 wbk.SaveAs(s,xlNormal,'','',false,false,xlNoChange,
           xlUserResolution,false,EmptyParam,EmptyParam,lcid);
end;

procedure TARExcelProvider.SaveDoc; 
begin
  wbk.Save(lcid);
end;

procedure TARExcelProvider.BeforeQuit; 
begin
  ea.CutCopyMode[lcid] := 0;
end;  

procedure TARExcelProvider.CloseDoc; 
begin
  wbk.Close(false,emptyparam,emptyparam,lcid);
end;

procedure TARExcelProvider.MarkSaved; 
begin
  wbk.Saved[lcid] := true;
end;
           
procedure TARExcelProvider.Quit;
begin
  ws := nil;
  wbk := nil;
  ea.quit;
  ea.Disconnect;
  Freeandnil(ea);
end;

function TARExcelProvider.GetDate(val : TDateTime): OLEVariant;
begin
   Result := Val;
end;

function TARExcelProvider.GetMyVariant(Field: TField;str:boolean): OLEVariant;
begin
   Result := '';
  case Field.DataType of
    ftString, ftFixedChar, ftWideString, ftMemo, ftFmtMemo:
    if str then
     Result := '''' + Field.AsString
    else
     Result := Field.AsString;
    ftSmallint, ftInteger, ftWord, ftLargeint, ftAutoInc: Result := Field.AsInteger;
    ftFloat, ftCurrency, ftBCD: Result := Field.AsFloat;
    ftBoolean: Result := Field.AsBoolean;
    ftDate, ftTime, ftDateTime: Result := Field.AsDateTime;
  end;
end;

function TARExcelProvider.DocLoaded : boolean;
begin
 result := assigned(wbk);
end;

procedure TARExcelProvider.Group(row, count : integer);
begin
 ws.Range['A'+inttostr(row),'A'+inttostr(row+count - 1)].Rows.Group(emptyparam,emptyparam,emptyparam,emptyparam);
end;


end.